@@ -23,10 +23,10 @@
 
 .rs.addFunction("subsetData", function(data, maxRows = -1, maxCols = -1)
 {
-   if (maxRows != -1 && nrow(data) > maxRows)
+   if (!is.na(maxRows) && maxRows != -1 && nrow(data) > maxRows)
       data <- head(data, n = maxRows)
    
-   if (maxCols != -1 && ncol(data) > maxCols)
+   if (!is.na(maxCols) && maxCols != -1 && ncol(data) > maxCols)
       data <- data[1:maxCols]
    
    data
@@ -100,15 +100,31 @@
 
 .rs.addFunction("formatDataColumnList", function(col, ...)
 {
-   limit <- .rs.readUserPref("data_viewer_max_cell_size")
-   if (is.null(limit)) {
-      limit <- 50L
+   limit <- .rs.nullCoalesce(
+      .rs.readUserPref("data_viewer_max_cell_size"),
+      50L
+   )
+   
+   # handle data.frame entries in a column
+   # https://github.com/rstudio/rstudio/issues/14257
+   for (i in seq_along(col))
+   {
+      if (is.data.frame(col[[i]]))
+      {
+         col[[i]] <- sprintf("<data.frame[%i x %i]>", nrow(col[[i]]), ncol(col[[i]]))
+      }
+      else
+      {
+         col[[i]] <- format(col[[i]])
+      }
    }
-
+   
    formatted <- as.character(col)
-   large <- nchar(formatted) > limit
+   na <- is.na(formatted)
+   large <- !na & nchar(formatted) > limit
    formatted <- substr(formatted, 1, limit)
    formatted <- paste0(formatted, ifelse(large, " [...]", ""))
+   formatted[na] <- NA_character_
    formatted
 })
 
@@ -138,7 +154,7 @@
    # we pass totalCols in the rownames col so we can pass this information
    # along when we retrieve column data, without changing the response format
    totalCols <- if (totalCols > 0) totalCols else ncol(x)
-
+   
    # the first column is always the row names
    rowNameCol <- list(
       col_name        = .rs.scalar(""),
@@ -177,10 +193,17 @@
       }
       else if (idx <= length(colLabels))
       {
-         if (col_name %in% names(colLabels))
-            colLabels[[col_name]]
+         if (!is.null(names(colLabels)))
+         {
+            if (col_name %in% names(colLabels))
+               colLabels[[col_name]]
+            else
+               ""
+         }
          else
+         {
             colLabels[[idx]]
+         }
       }
       else
       {
@@ -274,39 +297,84 @@
                                              sliceEnd = 1)
 {
    totalCols <- ncol(x)
-
    if (totalCols == 0)
       return(NULL)
-  
+   
    if (sliceEnd > totalCols || sliceEnd < 1)
       sliceEnd <- totalCols
+   
    if (sliceStart > totalCols || sliceStart < 1 || sliceStart > sliceEnd)
       sliceStart <- 1
    
-   colSlice <- x[sliceStart:sliceEnd]
+   indices <- sliceStart:sliceEnd
+   colSlice <- x[indices]
+   
+   # Make sure we preserve variable.labels if set.
+   #
+   # The structure of 'variable.labels' is not documented,
+   # but it appears that the expectation is that it's a character
+   # vector of the same length as 'x'.
+   #
+   # The vector can be optionally named (with names matching that of 'x'),
+   # or it can be an unnamed vector -- in which case, the order of labels
+   # needs to match the column order of 'x'.
+   #
+   # https://github.com/rstudio/rstudio/issues/14265
+   colLabels <- attr(x, "variable.labels", exact = TRUE)
+   if (!is.null(colLabels))
+   {
+      # Only subset 'variable.labels' if it's not named, since a named
+      # attribute could potentially be in a different order than the
+      # columns of 'x' itself. If 'variable.labels' is named, then all
+      # we require is that it's a super-set of the names of 'x'.
+      if (is.null(names(colLabels)))
+         colLabels <- colLabels[indices]
+      
+      attr(colSlice, "variable.labels") <- colLabels
+   }
+      
+   
    .rs.describeCols(colSlice, -1, -1, 64, totalCols)
 })
 
 .rs.addFunction("formatRowNames", function(x, start, len) 
 {
-   # detect whether this is a data.frame that contains
-   # row names, or if the row names are stored compactly
-   if (is.data.frame(x))
+   # check for a data.frame with compact row names
+   if (.rs.hasCompactRowNames(x))
    {
+      # the second element indicates the number of rows, and
+      # is negative if they're so-called "automatic" row names
       info <- .row_names_info(x, type = 0L)
-      if (is.integer(info) && length(info) > 0 && is.na(info[[1]]))
-      {
-         # the second element indicates the number of rows, and is negative if they're 
-         # automatic
-         n <- abs(info[[2]])
-         range <- seq(from = start, to = min(n, start + len))
-         return(as.character(range))
-      }
+      n <- abs(info[[2L]])
+      range <- seq(from = start, to = min(n, start + len))
+      return(as.character(range))
    }
    
-   # otherwise, extract row names and subset as usual
-   rownames <- row.names(x)
-   rownames[start:min(length(rownames), start + len)]
+   # retrieve row names; use .row_names_info for data.frame so
+   # we can detect internal non-character row names
+   rowNames <- if (is.data.frame(x))
+   {
+      .row_names_info(x, type = 0L)
+   }
+   else
+   {
+      row.names(x)
+   }
+   
+   # subset the retrieved row names
+   rowNames <- rowNames[start:min(length(rowNames), start + len)]
+   
+   # encode strings as JSON to force quoting + handle escaping
+   # this also lets us differentiate numeric (automatic) row names
+   # from explicitly-set row names
+   if (is.character(rowNames))
+   {
+      .rs.mapChr(rowNames, .rs.toJSON, unbox = TRUE)
+   }
+   else
+   {
+      as.character(rowNames)
+   }
 })
 
 # wrappers for nrow/ncol which will report the class of object for which we
@@ -341,14 +409,20 @@
       cols
 })
 
-.rs.addFunction("toDataFrame", function(x, name, flatten) {
-   
-   # if we have a data.table, convert it to a data.frame explicitly
-   if (inherits(x, "data.table"))
-      x <- as.data.frame(x)
+.rs.addFunction("toDataFrame", function(x, name, flatten)
+{
+   # force a non-subclassed data.frame -- this is necessary to ensure
+   # that row names (or row numbers) are not dropped when subsetting
+   # data, since those row names are used when generating cell-specific
+   # callbacks (e.g. for viewing a cell of a list column)
+   if (is.data.frame(x))
+   {
+      class(x) <- "data.frame"
+   }
    
    # if it's not already a frame, coerce it to a frame
-   if (!is.data.frame(x)) {
+   if (!is.data.frame(x))
+   {
       frame <- NULL
       # attempt to coerce to a data frame--this can throw errors in the case
       # where we're watching a named object in an environment and the user
@@ -388,9 +462,9 @@
       if (is.null(names(x)))
          names(x) <- paste("V", seq_along(x), sep = "")
       
-      if (flatten) {
+      if (flatten)
          x <- .rs.flattenFrame(x)
-      } 
+      
       return(x)
    }
 })
@@ -400,55 +474,79 @@
    which(vapply(x, fun, TRUE))
 })
 
-.rs.addFunction("flattenFrame", function(x) {
-  multicols <- .rs.multiCols(x)
-  while (length(multicols) > 0) {
-    multicol <- multicols[1]
-    newcols <- ncol(x[[multicol]])
-    if (identical(newcols, 0)) 
-    {
-      # remove columns consisting of empty frames
-      x[[multicol]] <- NULL
-    }
-    else
-    {
-      cols <- x[[multicol]]
+# given a 'data.frame' containing columns which themselves have
+# multiple columns (e.g. matrices, data.frames), expand those columns
+# such that we have a 'data.frame' with the nested columns e
+.rs.addFunction("flattenFrame", function(x)
+{
+   # skip if we don't have any rectangular columns;
+   # in this case, we can return the data as-is
+   isRectangular <- vapply(x, function(column) {
+      is.data.frame(column) || is.matrix(column)
+   }, FUN.VALUE = logical(1))
+   
+   if (!any(isRectangular))
+      return(x)
+   
+   # split into separate data.frames
+   stack <- .rs.stack()
+   .rs.enumerate(x, .rs.dataViewer.flatten, stack = stack)
+   parts <- stack$data()
+   
+   # pull out pieces we need
+   keys <- vapply(parts, `[[`, "name", FUN.VALUE = "character")
+   vals <- lapply(parts, `[[`, "value")
+   
+   # turn it into a data.frame
+   names(vals) <- keys
+   attr(vals, "row.names") <- .set_row_names(length(vals[[1L]]))
+   class(vals) <- "data.frame"
+   
+   # all done
+   vals
+})
 
-      # recurse because x[[multicol]] might also need flattening
-      if (length(.rs.multiCols(cols)) > 0L) {
-        cols <- x[[multicol]] <- .rs.flattenFrame(cols)
-        
-        # readjust indices
-        newcols <- ncol(cols)
-      }
-      
-      # apply column names
-      prefix <- names(multicol)[[1]]
-      if (is.matrix(cols)) {
-        colnames <- colnames(cols)
-        if (is.null(colnames)) {
-          colnames(cols) <- paste0(prefix, '[,', 1:ncol(cols), ']')
-        } else {
-          colnames(cols) <- paste0(prefix, '[,"', colnames, '"]')
-        }
+.rs.addFunction("dataViewer.flatten", function(name, value, stack)
+{
+   if (is.matrix(value)) {
+      .rs.dataViewer.flattenMatrix(name, value, stack)
+   } else if (is.data.frame(value)) {
+      .rs.dataViewer.flattenDataFrame(name, value, stack)
+   } else {
+      stack$push(list(name = name, value = value))
+   }
+})
 
-      } else if (is.data.frame(cols)) {
-        names(cols) <- paste(prefix, names(cols), sep = "$")
-      }
-      
-      # replace other columns in place
-      if (multicol >= ncol(x))  {
-        x <- cbind(x[0:(multicol-1)], cols)
-      } else {
-        x <- cbind(x[0:(multicol-1)], cols, x[(multicol+1):ncol(x)])
-      }
-    }
-    
-    # pop this frame off the list and adjust the other indices to account for
-    # the columns we just added, if any
-    multicols <- multicols[-1] + (max(newcols, 1) - 1) 
-  }
-  x
+.rs.addFunction("dataViewer.flattenMatrix", function(name, value, stack)
+{
+   colNames <- if (is.null(colnames(value)))
+      as.character(seq_len(ncol(value)))
+   else
+      encodeString(colnames(value), quote = '"')
+   
+   for (i in seq_len(ncol(value))) {
+      .rs.dataViewer.flatten(
+         name  = sprintf("%s[, %s]", name, colNames[[i]]),
+         value = value[, i, drop = TRUE],
+         stack = stack
+      )
+   }
+})
+
+.rs.addFunction("dataViewer.flattenDataFrame", function(name, value, stack)
+{
+   # a data.frame should almost always have names, but check just in case
+   colNames <- names(value)
+   if (is.null(colNames))
+      colNames <- sprintf("<%i>", seq_along(value))
+   
+   for (i in seq_along(value)) {
+      .rs.dataViewer.flatten(
+         name  = paste(name, colNames[[i]], sep = "$"),
+         value = value[[i]],
+         stack = stack
+      )
+   }
 })
 
 .rs.addFunction("applyTransform", function(x, filtered, search, cols, dirs)
@@ -459,19 +557,25 @@
          Encoding(colfilter) <- "UTF-8"
       colfilter
    }, "")
+   
    if (Encoding(search) == "unknown")
       Encoding(search) <- "UTF-8"
    
    # coerce argument to data frame--data.table objects (for example) report that
    # they're data frames, but don't actually support the subsetting operations
    # needed for search/sort/filter without an explicit cast
+   #
+   # similarly, we need to convert tibbles to regular data.frames so that we can
+   # properly invoke the list / data viewer on filtered rows
    x <- .rs.toDataFrame(x, "transformed", TRUE)
    
    # apply columnwise filters
-   for (i in seq_along(filtered)) {
-      if (nchar(filtered[i]) > 0 && length(x[[i]]) > 0) {
+   for (i in seq_along(filtered))
+   {
+      if (nchar(filtered[i]) > 0 && length(x[[i]]) > 0)
+      {
          # split filter--string format is "type|value" (e.g. "numeric|12-25") 
-         filter <- strsplit(filtered[i], split="|", fixed = TRUE)[[1]]
+         filter <- strsplit(filtered[i], split = "|", fixed = TRUE)[[1]]
          if (length(filter) < 2) 
          {
             # no filter type information
@@ -522,10 +626,31 @@
    # apply global search
    if (!is.null(search) && nchar(search) > 0)
    {
-      x <- x[Reduce("|", lapply(x, function(column) { 
-         grepl(paste("\\Q", search, "\\E", sep = ""), column, perl = TRUE,
-               ignore.case = TRUE)
-      })), , drop = FALSE]
+      # get columns for search
+      searchColumns <- unclass(x)
+      
+      # also apply on row names if available
+      if (is.data.frame(x))
+      {
+         info <- .row_names_info(x, type = 0L)
+         if (is.character(info))
+         {
+            searchColumns[[length(searchColumns) + 1]] <- info
+         }
+      }
+      
+      # apply global search on data columns
+      pattern <- paste0("\\Q", search, "\\E")
+      matches <- lapply(searchColumns, function(column) {
+         grepl(pattern, column, perl = TRUE, ignore.case = TRUE)
+      })
+      
+      # collapse into single vector
+      matches <- Reduce(`|`, matches)
+      
+      # update based on matches
+      x <- x[matches, , drop = FALSE]
+      
    }
    
    # apply sort
@@ -628,7 +753,7 @@
             return(get(cacheKey, envir = .rs.CachedDataEnv, inherits = FALSE))
       }
    }
-
+   
    # failure
    return(NULL)
 })
@@ -640,9 +765,9 @@
    # default to searching from the global environment
    env <- globalenv()
    
-   # attempt to find a callframe from which View was invoked; this will allow
-   # us to locate viewing environments further in the callstack (e.g. in the
-   # debugger)
+   # attempt to find a call frame from which View was invoked; this will allow
+   # us to locate viewing environments further in the call stack
+   # (e.g. in the debugger)
    for (i in seq_along(sys.calls()))
    {
       if (identical(deparse(sys.call(i)[[1]]), "View"))
@@ -652,17 +777,22 @@
       }
    }
    
-   while (environmentName(env) != "R_EmptyEnv" && 
-          !exists(name, where = env, inherits = FALSE)) 
-   {
-      env <- parent.env(env)
-   }
+   # NOTE: we previously looked through the parent environments of
+   # the associated frame to find the actual environment hosting the
+   # object being viewed, but this caused problems when attempting
+   # to track object mutations. for example, the 'mtcars' dataset is
+   # defined in 'package:datasets', but attempting to modify that
+   # object would actually create that modified object in the R global
+   # environment. for that reason, it's best to track objects from the
+   # top-level environment where they were found, as that's where
+   # "modified" versions of that object will be generated
    env
 })
 
 # attempts to determine whether the View(...) function the user has an 
 # override (i.e. it's not the handler RStudio uses)
-.rs.addFunction("isViewOverride", function() {
+.rs.addFunction("isViewOverride", function()
+{
    # check to see if View has been overridden: find the View() call in the 
    # stack and examine the function being evaluated there
    for (i in seq_along(sys.calls()))
@@ -674,6 +804,7 @@
          return(!identical(deparse(body(sys.function(i))[[1]]), ".rs.callAs"))
       }
    }
+   
    # if we can't find View on the callstack, presume the common case (no
    # override)
    FALSE
@@ -781,7 +912,7 @@
    
    if (!.rs.isNonEmptyScalarString(cacheKey))
       return(invisible(NULL))
-
+   
    # call viewData 
    invisible(.Call("rs_viewData", x, expr, title, name, env, cacheKey, FALSE))
 })
@@ -792,7 +923,7 @@
 {
    if (inherits(object, c("function", "vignette")))
       return(FALSE)
-
+   
    # prefer data viewer for pandas DataFrames
    if (inherits(object, "pandas.core.frame.DataFrame"))
       return(FALSE)
@@ -853,7 +984,7 @@
    {
       # remove data from the cache environment
       if (exists(".rs.CachedDataEnv") &&
-         exists(cacheKey, where = .rs.CachedDataEnv, inherits = FALSE))
+          exists(cacheKey, where = .rs.CachedDataEnv, inherits = FALSE))
          rm(list = cacheKey, envir = .rs.CachedDataEnv, inherits = FALSE)
       
       # remove data from the cache directory
@@ -881,8 +1012,8 @@
    lapply(ls(.rs.CachedDataEnv), function(cacheKey) {
       if (.rs.isNonEmptyScalarString(cacheKey))
          save(list = cacheKey, 
-            file = file.path(cacheDir, paste(cacheKey, "Rdata", sep = ".")),
-            envir = .rs.CachedDataEnv)
+              file = file.path(cacheDir, paste(cacheKey, "Rdata", sep = ".")),
+              envir = .rs.CachedDataEnv)
    })
    
    # clean the cache environment
@@ -930,6 +1061,6 @@
 
 .rs.addFunction("isNonEmptyScalarString", function(x)
 {
-   !is.language(x) && is.character(x) && length(x) == 1 && nzchar(x)
+   is.character(x) && length(x) == 1 && nzchar(x)
 })
 
